import os
import time

from agent.common.config import Config
from agent.common.system import sysCommand
from collections import defaultdict

from agent.common.pylog import logger

LIMITS_CONF = "/etc/security/limits.conf"
FILE_MAX = "/proc/sys/fs/file-max"


class Limits:
    def __init__(self):
        super().__init__()
        self.backup_file = os.path.join(Config.BACKUP_PATH, "limits.sh")


    def _getLimitsCmd(self, param_name, param_value):
        param = " ".join(param_name.split("_"))
        other_user = "* {}".format(param)
        root_user = "root {}".format(param)
        base_cmd = 'sed -i "/{0}/d" /etc/security/limits.conf;echo "{0} {1}" >> /etc/security/limits.conf'
        root_cmd = base_cmd.format(root_user, param_value)
        other_cmd = base_cmd.format(other_user, param_value)
        return """{};{}""".format(root_cmd, other_cmd)


    def _setLimitsFiles(self, param_name, param_value):
        _value = int(param_value)
        if "nofile" in param_name:
            cmd = self._getLimitsCmd(param_name, _value)
        elif param_name == "ulimit":
            cmd = "ulimit -n {}".format(_value)
        elif param_name == "file-max":
            cmd = "echo {} > {}".format(_value, FILE_MAX)
        return sysCommand(cmd)


    def _getParam(self, param_name):
        if "nofile" in param_name:
            param = " ".join(param_name.split("_"))
            cmd = "cat /etc/security/limits.conf |grep 'root {}'|awk -F' ' '{print $4}'".format(param)
        elif param_name == "ulimit":
            cmd = "ulimit -n"
        elif param_name == "file-max":
            cmd = 'cat {}'.format(FILE_MAX)
        return sysCommand(cmd)


    def setParamAll(self, param_list: dict):
        result = defaultdict(dict)
        success = False
        for param_name, param_info in param_list.items():
            _suc, _res = self._setLimitsFiles(param_name, param_info['value'])
            result[param_name] = {
                "value" : param_info['value'],
                "dtype" : param_info["dtype"],
                "suc"   : _suc,
                "msg"   : _res
            }
            success = success | _suc
        return success, result


    def getParamAll(self, param_list: dict):
        result = defaultdict(dict)
        for param_name, param_info in param_list.items():
            suc, param_value = self._getParam(param_name)
            result[param_name] = {
                "value" : param_value,
                "dtype" : param_info["dtype"],
                "suc"   : suc,
                "msg"   : ""
            }
        return True, result


    def rollback(self):
        if not os.path.exists(self.backup_file):
            return True, "Can not find backup file:{}".format(self.backup_file)

        suc, res = sysCommand("echo y|cp {} {}".format(
            os.path.join(Config.BACKUP_PATH, "limits.conf"), LIMITS_CONF))
        if suc:
            os.remove(os.path.join(Config.BACKUP_PATH, "limits.conf"))
        else:
            return False, "rollback limits conf failed: {}".format(res)

        suc, res = sysCommand("bash {}".format(self.backup_file))
        if suc:
            os.remove(self.backup_file)
        else:
            return False, "rollback limits conf failed: {}".format(res)
        
        return True, "rollback limits successfully!" 


    def backup(self, param_list: dict):
        suc, res = self.rollback()
        if not suc:
            return None
        
        backup_script = ""
        avaliable_knobs = defaultdict(dict)

        for param_name, _ in param_list.items():
            # Backup hard_nofile and hard_nofile by copy limits.conf
            if param_name in ["hard_nofile", "soft_nofile"]:
                suc, res = sysCommand("echo y|cp {} {}".format(
                    LIMITS_CONF, 
                    os.path.join(Config.BACKUP_PATH, "limits.conf")))
                if suc:
                    avaliable_knobs[param_name]['avaliable'] = True
                    avaliable_knobs[param_name]['value']     = None
                    avaliable_knobs[param_name]['msg']       = ""
                else:
                    avaliable_knobs[param_name]['avaliable'] = False
                    avaliable_knobs[param_name]['value']     = None
                    avaliable_knobs[param_name]['msg']       = res

            # Backup parameter ulimit
            elif param_name == "ulimit":
                suc, res = sysCommand("ulimit -n")
                if suc:
                    backup_script += "ulimit -n {}".format(res) + '\n'
                    avaliable_knobs[param_name]['avaliable'] = True
                    avaliable_knobs[param_name]['value']     = res
                    avaliable_knobs[param_name]['msg']       = ""
                else:
                    avaliable_knobs[param_name]['avaliable'] = False
                    avaliable_knobs[param_name]['value']     = None
                    avaliable_knobs[param_name]['msg']       = res

            # Backup parameter file-max
            elif param_name == "file-max":
                suc, res = sysCommand("cat {}".format(FILE_MAX))
                if suc:
                    backup_script += "echo {} > {}".format(res, FILE_MAX) + '\n'
                    avaliable_knobs[param_name]['avaliable'] = True
                    avaliable_knobs[param_name]['value']     = res
                    avaliable_knobs[param_name]['msg']       = ""
                else:
                    avaliable_knobs[param_name]['avaliable'] = False
                    avaliable_knobs[param_name]['value']     = None
                    avaliable_knobs[param_name]['msg']       = res

        with open(self.backup_file, 'w') as f:
            f.write(backup_script)
        
        return avaliable_knobs


if __name__ == "__main__":
    limits = Limits()
    # print(limits.setParamAll({"hard_nofile": {'value': 1048576, 'dtype': 'int'},
    #                           "soft_nofile": {'value': 1048576, 'dtype': 'int'},
    #                           "ulimit": {'value': 1048576, 'dtype': 'int'},
    #                           "file-max": {'value': 1048576, 'dtype': 'int'}}))
    print(limits.backup({"hard_nofile": {'value': 1048576, 'dtype': 'int'},
                              "soft_nofile": {'value': 1048576, 'dtype': 'int'},
                              "ulimit": {'value': 1048576, 'dtype': 'int'},
                              "file-max": {'value': 1048576, 'dtype': 'int'}}))